use cgmath::{vec2, vec3, Vector3};

use image::ImageBuffer;
use utilities::prelude::*;

mod aabb;
mod acceleration_data;
mod camera;
mod intersection;
mod ray;
mod triangle;

use acceleration_data::AccelerationData;
use camera::Camera;
use ray::Ray;
use triangle::{Triangle, Vertex};

fn generate_grid(width: u32, height: u32) -> Vec<Triangle> {
    let mut grid = Vec::new();

    for y in 0..height {
        for x in 0..width {
            let color = vec3(1.0, 0.0, 0.0);

            grid.push(Triangle::new(
                Vertex::new(vec3(x as f32, y as f32, 0.0), vec2(0.0, 0.0)),
                Vertex::new(vec3(x as f32 + 1.0, y as f32, 0.0), vec2(1.0, 0.0)),
                Vertex::new(vec3(x as f32 + 1.0, y as f32 + 1.0, 0.0), vec2(1.0, 1.0)),
                color,
            ));

            grid.push(Triangle::new(
                Vertex::new(vec3(x as f32 + 1.0, y as f32 + 1.0, 0.0), vec2(1.0, 1.0)),
                Vertex::new(vec3(x as f32, y as f32 + 1.0, 0.0), vec2(0.0, 1.0)),
                Vertex::new(vec3(x as f32, y as f32, 0.0), vec2(0.0, 0.0)),
                color,
            ));
        }
    }

    grid
}

fn main() {
    let input_data = generate_grid(3, 3);

    /*
    let camera = Camera::new(
        vec3(0.0, -4.0, 4.0),
        vec3(0.0, 0.0, 0.0),
        vec3(0.0, 0.0, 1.0),
        45.0,
        1280.0 / 720.0,
    );
    */

    let camera = Camera::new(
        vec3(0.0, 0.0, 3.0),
        vec3(0.0, 0.0, 0.0),
        vec3(0.0, 1.0, 1.0),
        45.0,
        1280.0 / 720.0,
    );

    debug_raytracer_camera(1280, 720, &camera, &input_data);
}

fn f_to_u(color: f32) -> u8 {
    ((color * 255.0) / 1.0) as u8
}

fn debug_raytracer_camera(dim_x: u32, dim_y: u32, camera: &Camera, data: &[Triangle]) {
    let mut imgbuf = ImageBuffer::new(dim_x, dim_y);

    let acceleration_data = AccelerationData::create_tree(data, 8);

    for (x, y, pixel) in imgbuf.enumerate_pixels_mut() {
        let ray = camera.primary_ray(x, dim_y - y, dim_x, dim_y);

        let color = pixel_color(&ray, &acceleration_data);

        *pixel = image::Rgb([f_to_u(color.x), f_to_u(color.y), f_to_u(color.z)]);
    }

    imgbuf.save("raytrace.png").unwrap();
}

fn pixel_color(ray: &Ray, acceleration_data: &AccelerationData) -> Vector3<f32> {
    let mut final_color = vec3(0.0, 1.0, 0.0);

    /*
    let mut closest_value = MAX;
    let mut closest_index = -1;
    let mut barycentric = Vector2::zero();

    for accel in acceleration_data.iter() {
        if accel.intersect(ray) {
            for i in accel.start_index..accel.end_index {
                let mut t = 0.0;
                let mut tmp_bary = Vector2::zero();

                if data[i].intersect_mt(ray, &mut t, &mut tmp_bary) {
                    if t < closest_value {
                        closest_index = i as i32;
                        closest_value = t;
                        barycentric = tmp_bary;
                    }
                }
            }
        }
    }
    */

    if let Some(intersection) = acceleration_data.find_closest_hit(ray) {
        let triangle = intersection.triangle;

        /*
        let wuv = vec3(
            1.0 - intersection.barycentric.x - intersection.barycentric.y,
            intersection.barycentric.x,
            intersection.barycentric.y,
        );

        let st = point_st(&triangle, &wuv);
        */

        final_color = triangle.color;
    }

    return final_color;
}

/*
fn point_st(triangle: &Triangle, wuv: &Vector3<f32>) -> Vector2<f32> {
    return wuv.x * triangle.vertices[0].texture_coordinate
        + wuv.y * triangle.vertices[1].texture_coordinate
        + wuv.z * triangle.vertices[2].texture_coordinate;
}
*/
