use cgmath::{InnerSpace, Vector2, Vector3, Zero};

use super::ray::Ray;

#[derive(Clone, Copy)]
pub struct Vertex {
    pub position: Vector3<f32>,
    pub texture_coordinate: Vector2<f32>,
}

impl Vertex {
    pub fn new(position: Vector3<f32>, texture_coordinate: Vector2<f32>) -> Vertex {
        Vertex {
            position,
            texture_coordinate,
        }
    }

    pub fn zero() -> Vertex {
        Vertex {
            position: Vector3::zero(),
            texture_coordinate: Vector2::zero(),
        }
    }
}

pub struct Triangle {
    pub vertices: [Vertex; 3],
    pub color: Vector3<f32>,
}

impl Triangle {
    pub fn new(v0: Vertex, v1: Vertex, v2: Vertex, color: Vector3<f32>) -> Triangle {
        Triangle {
            vertices: [v0, v1, v2],
            color,
        }
    }

    pub fn zero() -> Triangle {
        Triangle {
            vertices: [Vertex::zero(); 3],
            color: Vector3::zero(),
        }
    }

    // source: https://www.scratchapixel.com/lessons/3d-basic-rendering/ray-tracing-rendering-a-triangle/moller-trumbore-ray-triangle-intersection
    pub fn intersect_mt(&self, ray: &Ray, t: &mut f32, barycentric: &mut Vector2<f32>) -> bool {
        let v0v1 = self.vertices[1].position - self.vertices[0].position;
        let v0v2 = self.vertices[2].position - self.vertices[0].position;
        let pvec = ray.direction.cross(v0v2);
        let det = v0v1.dot(pvec);

        // ray and triangle are parallel if det is close to 0
        if det.abs() < 0.001 {
            return false;
        }

        let inv_det = 1.0 / det;

        let tvec = ray.origin - self.vertices[0].position;
        barycentric.x = tvec.dot(pvec) * inv_det;

        if barycentric.x < 0.0 || barycentric.x > 1.0 {
            return false;
        }

        let qvec = tvec.cross(v0v1);
        barycentric.y = ray.direction.dot(qvec) * inv_det;

        if barycentric.y < 0.0 || barycentric.x + barycentric.y > 1.0 {
            return false;
        }

        *t = v0v2.dot(qvec) * inv_det;

        true
    }
}
